package com.unitesk.requality.core.instancer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;

import org.mozilla.javascript.Context;
import org.mozilla.javascript.NativeObject;
import org.mozilla.javascript.ScriptRuntime;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.WrapFactory;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeStorage;

public class TreeInstancer extends TreeStorage {
    private TreeStorage storage;
    private TreeDB srcTree;
    private Map<UUID, String> nodePredicates = new HashMap<UUID, String>();
    private Map<UUID, Collection<UUID>> toChildren = new LinkedHashMap<UUID, Collection<UUID>>();
    private Map<UUID, Map<String,String>> nodeAttribute = new HashMap<UUID, Map<String,String>>();

    public TreeInstancer(TreeDB db) {
        this.storage = db.getStorage();
        this.srcTree = db;
    }

    @Override
    public UUID getRootNode() {
        if(!nodeAttribute.containsKey(storage.getRootNode())) {
            nodeAttribute.put(storage.getRootNode(), getAttributesMap(storage.getRootNode()));
        }
        if(!nodePredicates.containsKey(storage.getRootNode())) {
            nodePredicates.put(storage.getRootNode(), getPredicate(storage.getRootNode()));
        }
        return storage.getRootNode();
    }

    @Override
    public UUID getParent(UUID uuid) {
        return storage.getParent(uuid);
    }

    @Override
    public Collection<UUID> getChildren(UUID uuid) {
        if(toChildren.containsKey(uuid)) {
            return toChildren.get(uuid);
        }
        Collection<UUID> result = new ArrayList<UUID>(storage.getChildren(uuid));
        for(String qid : srcTree.getNode(uuid).getBaseNodes()) {
            result.addAll(getChildren(srcTree.getNode(qid).getUUId()));
        }
        
        Collection<UUID> tmp = new ArrayList<UUID>(result);
        for(UUID id: tmp) {
            if(!checkPredicate(id)) {
                result.remove(id);
            }
        }
        toChildren.put(uuid, result);
        return result;
    }

    private boolean checkPredicate(UUID id) {
        String predicate = getPredicate(id);
        boolean result = true;
        result = evaluatePredicate(predicate, getAttributesMap(id));
        // TODO: remove
        System.out.println("(node '" + getNodeId(id) + "'): "+predicate + " is " + result);
        return result;
    }

    public String getPredicate(UUID id) {
        if(nodePredicates.containsKey(id)) {
            return nodePredicates.get(id);
        }
        if(id.equals(storage.getRootNode())) {
            return srcTree.getNode(id).getPredicate();
        }

        String full = "";
        if(getParent(id)!=null) {
            full = getPredicate(getParent(id));
        }
        String newPart = srcTree.getNode(id).getPredicate();

        return concatPredicate(full, newPart);
    }
    
    private String concatPredicate(String full, String newPart) {
        if(full.length()==0) return newPart;
        if(newPart.length()==0)  return full;
            
        if(full.equals(newPart)) return full;
        return "("+full+") && ("+newPart+")";
    }
    
    private Map<String, String> getAttributesMap(UUID id) {
        if(nodeAttribute.containsKey(id)) {
            return nodeAttribute.get(id);
        }
        Map<String, String> result = new HashMap<String, String>(); 
        if(getParent(id) != null && !id.equals(storage.getRootNode())) {
            result.putAll(getAttributesMap(getParent(id)));
        }

        for(String key : srcTree.getNode(id).getAttributeKeys()) {
            result.put(key, srcTree.getNode(id).getAttributeValue(key).toString());
        }
        return result;
    }

    private boolean evaluatePredicate(String predicate, Map<String, String> attributes) {
        if(predicate==null || predicate.length()==0) return true;
        
       
        Boolean contextResult=true;
        try {
            Context context = Context.enter();
            WrapFactory wrapFactory = new WrapFactory();
            context.setWrapFactory(wrapFactory);
            Scriptable scope = createScope(context, attributes);
            contextResult = (Boolean) context.evaluateString( scope, predicate, "", 1, null );
        } finally {
            Context.exit();
        }
        return contextResult;
    }
    
    private Scriptable createScope(Context context, Map<String, String> attributes) {
        Scriptable topScope = ScriptRuntime.initStandardObjects(context, null, true);
        Scriptable scope = new NativeObject();
        scope.setParentScope( null );
        scope.setPrototype( topScope );

        for(String attrName: attributes.keySet() ) {
            String value = attributes.get(attrName);
            Object obj = context.getWrapFactory().wrap( context, scope, value, value.getClass() );
            scope.put( attrName, scope, obj );
        }

        return scope;
    }

    @Override
    public void createNode(UUID parentUUId, UUID uuid, String id,
            Collection<Attribute> attrs) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void remove(UUID node) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getNodeId(UUID uuid) {
        return storage.getNodeId(uuid);
    }

    @Override
    public void changeNodeId(UUID node, String id) {
        throw new UnsupportedOperationException();
    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
        return storage.getAttributes(uuid);
    }

    @Override
    public void saveAttributes(UUID node, Collection<Attribute> attributes) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void moveNodeTo(UUID node, UUID parent) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean canLoad(UUID uuid) {
        return storage.canLoad(uuid);
    }

    @Override
    public String getNodeResourcePath(UUID uuId) {
        return storage.getNodeResourcePath(uuId);
    }
}
