/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.views.actions.documents;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.views.actions.CNFAction;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class RemoveDocumentAction extends CNFAction {
    private static final String REMOVE_DOCUMENT = "Remove Document";
    private static final String REMOVE_FOLDER = "Remove Folder";
    private static final String REMOVE_LOCATION = "Remove Location";

    private Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
            .getShell();


    public RemoveDocumentAction(ISelectionProvider provider) {
//        super(provider);
        setImageDescriptor(PlatformUI.getWorkbench().getSharedImages()
                .getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));
    }

    public void run() {
        TreeNode node = getSelectedNode();
        if (node instanceof DocFolder) {
            recDelNode(node);
        }
        if (node instanceof Document) {
            removeDocument((Document) node);
        }
    }

    @Override
    public boolean isEnabled() {
        TreeNode node = getSelectedNode();
        if (node == null) {
            return false;
        }
        if (node.getQualifiedId().equals(Document.getTypeRootQId())) {
            return false;
        }
        if(! (node.getType().equals("Document") || node.getType().equals("DocFolder"))) {
            return false;
        }
        return true;
    }


    private void recDelNode(TreeNode node) {
        if (node instanceof Document) {
            removeDocument((Document) node);
        }

        if (node instanceof DocFolder) {
            if (!MessageDialog.openConfirm(shell, REMOVE_FOLDER, "Delete selected document folder?")) {
                return;
            }
            if (folderHasDocumentWithLocation((DocFolder) node)) {
                if (!MessageDialog
                        .openConfirm(shell, REMOVE_FOLDER,
                                "Folder contains document with marked requirements, continue?")) {
                    return;
                }
            }

            for(int i=0; i < node.sizeChildren(); i++) {
                recDelNode(node.getChild(i));
            }
            node.delete();
        }
    }

    private void removeDocument(Document doc) {

        if (!MessageDialog.openConfirm(shell, REMOVE_DOCUMENT,
                "Delete selected document?")) {
            return;
        }

        if (doc.sizeChildren() > 0) {
            if (!MessageDialog.openConfirm(shell, REMOVE_FOLDER,
                    "Document contains marked requirements, continue?")) {
                return;
            }
        }
        removeLocationsFromRequirementd(doc);
        doc.delete();
    }

    private boolean folderHasDocumentWithLocation(DocFolder folder) {
        if (folder.sizeChildren() == 0) {
            return false;
        }

        for (int i = 0; i < folder.sizeChildren(); i++) {
            TreeNode node = folder.getChild(i);
            if (node instanceof DocFolder) {
                if (folderHasDocumentWithLocation((DocFolder) node)) {
                    return true;
                }
            }
            if (node instanceof Document) {
                Document doc = (Document) node;
                if (doc.sizeLocations() > 0) {
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public String getText() {
        TreeNode node = getSelectedNode();
        if (node instanceof Document) {
            return REMOVE_DOCUMENT;
        }

        if (node instanceof DocFolder) {
            return REMOVE_FOLDER;
        }
        if (node instanceof Location) {
            return REMOVE_LOCATION;
        }
        return "Remove";
    }

    private void removeLocationsFromRequirementd(Document doc) {
        TreeDB db = doc.getTreeDB();
        Set<String> locQIds = new HashSet<String>();
        for(int i=0; i < doc.sizeChildren(); i++) {
            locQIds.add(doc.getChild(i).getQualifiedId());
        }
        
        for (TreeNode node : new DeepFirstTreeWalker(db.getNode(Requirement
                .getTypeRootQId()))) {

            if(!node.getType().equals("Requirement")) {
                continue;
            }

            ArrayList<Location> removeLocations = new ArrayList<Location>();
            Requirement req = (Requirement) node;

            for (Location loc : req.getLocations()) {
                if (locQIds.contains(loc.getQualifiedId())) {
                    removeLocations.add(loc);
                }
            }

            for (Location loc : removeLocations) {
                req.removeLocation(loc.getQualifiedId());
            }
        }
    }
}
